import math
from scipy.io.arff import loadarff
from weka.core.dataset import Instances, Instance, Attribute
from datetime import datetime
from weka.core.dataset import missing_value
from weka.core.converters import loader_for_file, Loader
import numpy as np
from numpy import ndarray


def parse_range(r, max_value, ordered=True, safe=True):
    """
    Parses the Weka range string (eg "first-last" or "1,3-5,7,10-last")
    of 1-based indices and returns a list of 0-based integers.
    'first' and 'last' are accepted apart from integer strings,
    '-' is used to define a range (low to high).
    The list can be returned ordered or as is.

    :param r: the range string to parse
    :type r: str
    :param max_value: the maximum value for the 1-based indices
    :type max_value: int
    :param ordered: whether to return the list ordered or as is
    :type ordered: bool
    :param safe: whether to catch exceptions or not
    :type safe: bool
    :return: the list of 0-base indices
    :rtype: list
    """
    result = []

    parts = r.replace(" ", "").replace("first", "1").replace("last", str(max_value)).split(",")
    for part in parts:
        if "-" in part:
            bounds = part.split("-")
            low = None
            high = None
            if len(bounds) == 2:
                if safe:
                    try:
                        low = int(bounds[0]) - 1
                        high = int(bounds[1]) - 1
                        if low > high:
                            print("Invalid format for index range ('LOW-HIGH'): %s" % part)
                            continue
                    except:
                        print("Failed to parse bounds of '%s' from range '%s' (max: %d), skipping!" % (part, r, max_value))
                else:
                    low = int(bounds[0]) - 1
                    high = int(bounds[1]) - 1
                    if low > high:
                        print("Invalid format for index range ('LOW-HIGH'): %s" % part)
                        continue
            if (low is not None) and (high is not None):
                for i in range(low, high + 1):
                    result.append(i)
        else:
            if safe:
                try:
                    result.append(int(part) - 1)
                except:
                    print("Failed to parse '%s' from range '%s', skipping (max: %d)!" % (part, r, max_value))
            else:
                result.append(int(part) - 1)

    if ordered:
        result.sort()

    return result


def to_nominal_attributes(X, indices):
    """
    Turns the specified indices numeric column vector into a string vector.

    :param X: the 2D matrix to convert
    :type X: ndarray
    :param indices: the list of 0-based indices of attributes to convert to nominal or range string with 1-based indices
    :type indices: list or str
    :return: the converted matrix
    :rtype: ndarray
    """
    result = []
    if isinstance(indices, str):
        indices = parse_range(indices, X.shape[1])
    indices_set = set(indices)
    for r in X:
        rn = []
        for i in range(len(r)):
            if i in indices_set:
                rn.append("_" + str(r[i]))
            else:
                rn.append(r[i])
        result.append(rn)

    return np.array(result)


def to_nominal_labels(y):
    """
    Turns the numeric column vector into a string vector.

    :param y: the vector to convert
    :type y: list or ndarray
    :return: the converted vector
    :rtype: ndarray
    """
    result = []
    for item in y:
        if isinstance(item, np.bytes_):
            result.append("_" + item.astype(str))
        else:
            result.append("_" + str(item))
    return result


def split_off_class(data, class_index):
    """
    Splits off the class attribute from the data matrix.
    The class index can either be a 0-based int or a 1-based string
    (first,second,last,last-1 are accepted as well).

    :param data: the 2D matrix to process
    :type data: ndarray
    :param class_index: the position of the class attribute to split off
    :type class_index: int or str
    :return: the input variables (2D matrix) and the output variable (1D)
    """

    if len(data) == 0:
        num_atts = -1
    else:
        num_atts = len(data[0])

    # interpret class index
    if isinstance(class_index, str):
        if class_index == "first":
            index = 0
        elif class_index == "second":
            index = 1
        elif class_index == "last-1":
            if num_atts == -1:
                raise Exception("No data, cannot determine # of attributes for class index: %s" % class_index)
            index = num_atts - 2
        elif class_index == "last":
            if num_atts == -1:
                raise Exception("No data, cannot determine # of attributes for class index: %s" % class_index)
            index = num_atts - 1
        else:
            try:
                index = int(class_index) - 1
            except:
                raise Exception("Unsupported class index: %s" % class_index)
    elif isinstance(class_index, int):
        index = class_index
    else:
        raise Exception("Unsupported type for class index: " + str(type(class_index)))

    if index == 0:
        X = [list(r)[1:] for r in data]
    elif index == (num_atts - 1):
        X = [list(r)[0:index] for r in data]
    else:
        X = []
        for r in data:
            r = list(r)
            rn = []
            rn.extend(r[0:index])
            rn.extend(r[index+1:])
            X.append(rn)
    y = [r[index] for r in data]
    return np.array(X), np.array(y)


def load_arff(fname, class_index=None):
    """
    Loads the specified ARFF file. If a class index is provided, either 0-based int or 1-based string
    (first,second,last,last-1 are accepted as well), then the data is split into input variables and
    class attribute.

    :param fname: the path of the ARFF file to load
    :type fname: str
    :param class_index: the class index, either int or str
    :type class_index: int or str
    :return: tuple (X, meta), or in case of a valid class index a tuple (X,y,meta)
    :rtype: tuple
    """
    data, meta = loadarff(fname)
    if class_index is None:
        X = np.array([list(r) for r in data])
        return np.array(X), meta
    else:
        X, y = split_off_class(data, class_index)
        return X, y, meta


def load_dataset(fname, loader=None, class_index=None, internal=False):
    """
    Loads the dataset using Weka's converters. If no loader instance is provided, the extension of
    the file is used to determine a loader (using default options). The data can either be returned
    using mixed types or just numeric (using Weka's internal representation).

    :param fname: the path of the dataset to load
    :type fname: str
    :param loader: the customized Loader instance to use for loading the dataset, can be None
    :type loader: Loader
    :param class_index: the class index string to use ('first', 'second', 'third', 'last-2', 'last-1', 'last' or 1-based index)
    :type class_index: str
    :param internal: whether to return Weka's internal format or mixed data types
    :type internal: bool
    :return: the dataset tuple: (X) if no class index; (X,y) if class index
    """
    if loader is None:
        loader = loader_for_file(fname)
    weka_ds = loader.load_file(fname, class_index=class_index)
    numpy_ds = weka_ds.to_numpy(internal=internal)
    if class_index is not None:
        return split_off_class(numpy_ds, class_index)
    else:
        return numpy_ds


def determine_attribute_types(X):
    """
    Determines the type of the columns.

    :param X: the 2D data to determine the column types for
    :type X: ndarray
    :return: the list of types (C=categorical, N=numeric)
    :rtype: list
    """
    if len(X) == 0:
        raise Exception("No data to convert!")

    num_rows = len(X)
    num_cols = len(X[0])

    # initialize types
    result = []
    for i in range(num_cols):
        result.append("N")

    for i in range(num_cols):
        for n in range(num_rows):
            r = X[n]
            try:
                float(r[i])
            except:
                result[i] = "C"
                break

    return result


def determine_attribute_type(y):
    """
    Determines the type of the column.

    :param y: the 1D vector to determine the type for
    :type y: ndarray
    :return: the type (C=categorical, N=numeric)
    :rtype: str
    """
    result = "N"
    for i in range(len(y)):
        try:
            float(y[i])
        except:
            result = "C"
            break

    return result


def to_instances(X, y=None, att_names=None, att_types=None, class_name=None, class_type=None, relation_name=None,
                 num_nominal_labels=None, num_class_labels=None):
    """
    Turns the 2D matrix and the optional 1D class vector into an Instances object.

    :param X: the input variables, 2D matrix
    :type X: ndarray
    :param y: the optional class value column, 1D vector
    :type y: ndarray
    :param att_names: the list of attribute names
    :type att_names: list
    :param att_types: the list of attribute types (C=categorical, N=numeric), assumes numeric by default if not provided
    :param class_name: the name of the class attribute
    :type class_name: str
    :param class_type: the type of the class attribute (C=categorical, N=numeric)
    :type class_type: str
    :param relation_name: the name for the dataset
    :type relation_name: str
    :param num_nominal_labels: the dictionary with the number of labels (key is 0-based attribute index)
    :type num_nominal_labels: dict
    :param num_class_labels: the number of labels in the class attribute
    :type num_class_labels: int
    :return: the generated Instances object
    :rtype: Instances
    """

    if len(X) == 0:
        raise Exception("No data to convert!")

    # defaults
    if att_types is None:
        att_types = determine_attribute_types(X)
    if att_names is None:
        att_names = []
        for i in range(len(X[0])):
            att_names.append("att-" + str(i+1))
    if relation_name is None:
        relation_name = "scikit-weka @ " + str(datetime.now())
    if class_name is None:
        if "class" not in att_names:
            class_name = "class"
        else:
            class_name = "class-" + str(len(att_names) + 1)
    if y is not None:
        if class_type is None:
            class_type = determine_attribute_type(y)

    # create header
    atts = []

    for i in range(len(X[0])):
        att_name = att_names[i]
        att_type = att_types[i]

        if att_type == "N":
            atts.append(Attribute.create_numeric(att_name))
        elif att_type == "C":
            if (num_nominal_labels is not None) and (i in num_nominal_labels):
                values = []
                for l in range(num_nominal_labels[i]):
                    values.append("_%d" % l)
            else:
                labels = set()
                for n in range(len(X)):
                    r = X[n]
                    v = str(r[i])
                    labels.add(v)
                values = sorted(labels)
            atts.append(Attribute.create_nominal(att_name, values))
        else:
            raise Exception("Unsupported attribute type for column %d: %s" % ((i+1), att_type))

    if y is not None:
        if class_type == "N":
            atts.append(Attribute.create_numeric(class_name))
        elif class_type == "C":
            if num_class_labels is not None:
                values = []
                for l in range(num_class_labels):
                    values.append("_%d" % l)
            else:
                values = sorted(set([str(x) for x in y]))
            atts.append(Attribute.create_nominal(class_name, values))

    result = Instances.create_instances(relation_name, atts, len(X))
    if y is not None:
        result.class_index = result.num_attributes - 1

    # data
    for n in range(len(X)):
        values = []
        r = X[n]
        for i in range(len(r)):
            if att_types[i] == "C":
                values.append(atts[i].index_of(str(r[i])))
            elif att_types[i] == "N":
                values.append(r[i])
            else:
                raise Exception("Unsupported attribute type for column %d: %s" % ((i+1), att_types[i]))
        if y is not None:
            if class_type == "C":
                values.append(atts[-1].index_of(str(y[n])))
            elif class_type == "N":
                values.append(y[n])
            else:
                raise Exception("Unsupported attribute type for class: %s" % class_type)
        inst = Instance.create_instance(values)
        result.add_instance(inst)

    return result


def to_instance(header, x, y=None, weight=1.0):
    """
    Generates an Instance from the data.

    :param header: the data structure to adhere to
    :type header: Instances
    :param x: the 1D vector with input variables
    :type x: ndarray
    :param y: the optional class value
    :type y: object
    :param weight: the weight for the Instance
    :type weight: float
    :return: the generate Instance
    :rtype: Instance
    """
    values = []

    for i in range(len(x)):
        if header.attribute(i).is_nominal:
            values.append(header.attribute(i).index_of(str(x[i])))
        elif header.attribute(i).is_numeric:
            values.append(x[i])
        else:
            raise Exception("Unsupported attribute type for column %d: %s" % ((i+1), header.attribute(i).type_str()))

    if y is not None and header.has_class():
        # missing value?
        if math.isnan(y):
            values.append(missing_value())
        elif header.class_attribute.is_nominal:
            values.append(header.class_attribute.index_of(str(y)))
        elif header.class_attribute.is_numeric:
            values.append(y)
        else:
            raise Exception("Unsupported attribute type for class attribute: %s" % header.class_attribute.type_str())

    result = Instance.create_instance(values, weight=weight)
    result.dataset = header
    return result


def to_array(data):
    """
    Turns the Instances object into ndarrays for X and y. If no class is present, then y will be None.

    :param data: the data to convert
    :type data: Instances
    :return: the generated arrays for X and y
    :rtype: tuple
    """
    has_class = data.has_class()
    class_index = data.class_index
    X = []
    y = [] if has_class else None
    for i in range(data.num_instances):
        inst = data.get_instance(i)
        row = []
        for n in range(data.num_attributes):
            if n == class_index:
                continue
            if data.attribute(n).is_numeric:
                row.append(inst.get_value(n))
            else:
                row.append(inst.get_string_value(n))
        X.append(row)
        if has_class:
            if data.class_attribute.is_numeric:
                y.append(inst.get_value(class_index))
            else:
                y.append(inst.get_string_value(class_index))
    X = np.array(X)
    if y is not None:
        y = np.array(y)
    return X, y



import numpy as np
from weka.datagenerators import DataGenerator


def generate_data(generator, att_names=False):
    """
    Generates data using the Weka data generator.

    :param generator: the data generator to use
    :type generator: DataGenerator
    :param att_names: whether to return the attribute names as well
    :type att_names: bool
    :return: tuple of X and y matrices or tuple of X, y, X_names, y_name
    :rtype: tuple
    """

    format = generator.define_data_format()
    data = generator.generate_examples()
    index = data.class_index
    X = []
    X_names = []
    y = [] if (index > -1) else None
    y_name = None
    att_types = []
    for i in range(data.num_attributes):
        if i == index:
            y_name = data.attribute(i).name
        else:
            X_names.append(data.attribute(i).name)
        if data.attribute(i).is_numeric:
            att_types.append("N")
        else:
            att_types.append("S")
    for inst in data:
        row = []
        for i in range(data.num_attributes):
            if att_types[i] == "N":
                value = inst.get_value(i)
            else:
                value = inst.get_string_value(i)
            if i == index:
                y.append(value)
            else:
                row.append(value)
        X.append(row)

    if att_names:
        if y is None:
            return np.asarray(X), None, X_names, None
        else:
            return np.asarray(X), np.asarray(y), X_names, y_name
    else:
        if y is None:
            return np.asarray(X), None
        else:
            return np.asarray(X), np.asarray(y)



import numpy as np
from numpy import ndarray
from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.utils.validation import check_is_fitted
from sklweka.preprocessing import to_nominal_attributes
from weka.clusterers import Clusterer
from weka.core.classes import is_instance_of, OptionHandler
from weka.core.serialization import deepcopy
from sklweka.dataset import to_instances, to_instance


class WekaCluster(BaseEstimator, OptionHandler, ClusterMixin):
    """
    Wraps a Weka cluster within the scikit-learn framework.
    """

    def __init__(self, jobject=None, cluster=None, classname=None, options=None, nominal_input_vars=None,
                 num_nominal_input_labels=None):
        """
        Initializes the estimator. Can be either instantiated via the following priority of parameters:
        1. JB_Object representing a Java Clusterer object
        2. Clusterer pww3 wrapper
        3. classname/options

        :param jobject: the JB_Object representing a Weka cluster to use
        :type jobject: JB_Object
        :param cluster: the cluster wrapper to use
        :type cluster: Clusterer
        :param classname: the classname of the Weka cluster to instantiate
        :type classname: str
        :param options: the command-line options of the Weka cluster to instantiate
        :type options: list
        :param num_nominal_input_labels: the dictionary with the number of labels for the nominal input variables (key is 0-based attribute index)
        :type num_nominal_input_labels: dict
        """
        if jobject is not None:
            _jobject = jobject
        elif cluster is not None:
            _jobject = cluster.jobject
        elif classname is not None:
            if options is None:
                options = []
            cluster = Clusterer(classname=classname, options=options)
            _jobject = cluster.jobject
        else:
            raise Exception("At least Java classname must be provided!")

        if not is_instance_of(_jobject, "weka.clusterers.Clusterer"):
            raise Exception("Java object does not implement weka.clusterers.Clusterer!")

        super(WekaCluster, self).__init__(_jobject)
        self._cluster = Clusterer(jobject=_jobject)
        self.header_ = None
        # the following references are required for get_params/set_params
        self._classname = classname if (classname is not None) else self._cluster.classname
        self._options = options if (options is not None) else self._cluster.options
        self._nominal_input_vars = nominal_input_vars
        self._num_nominal_input_labels = num_nominal_input_labels

    @property
    def cluster(self):
        """
        Returns the underlying cluster object, if any.

        :return: the cluster object
        :rtype: Clusterer
        """
        return self._cluster

    @property
    def header(self):
        """
        Returns the underlying dataset header, if any.

        :return: the dataset structure
        :rtype: Instances
        """
        return self.header_

    def fit(self, X, y=None, sample_weight=None):
        """
        Trains the cluster.

        :param X: the input variables as matrix, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: ignored
        :type y: ndarray
        :param sample_weight: Sample weights. If None, then samples are equally weighted. TODO Currently ignored.
        :type sample_weight: array-like of shape (n_samples,), default=None
        :return: the cluster
        :rtype: WekaCluster
        """
        if self._nominal_input_vars is not None:
            X = to_nominal_attributes(X, self._nominal_input_vars)
        d = to_instances(X, num_nominal_labels=self._num_nominal_input_labels)
        self._cluster.build_clusterer(d)
        self.header_ = d.template_instances(d, 0)
        return self

    def predict(self, X, y=None):
        """
        Predicts cluster labels.

        :param X: the input variables as matrix, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: ignored
        :type y: ndarray
        :return: the cluster labels (of type int)
        :rtype: ndarray
        """
        check_is_fitted(self)
        if self._nominal_input_vars is not None:
            X = to_nominal_attributes(X, self._nominal_input_vars)
        result = []
        for d in X:
            inst = to_instance(self.header_, d)
            result.append(int(self._cluster.cluster_instance(inst)))
        return np.array(result)

    def fit_predict(self, X, y=None, sample_weight=None):
        """
        Trains the cluster and returns the cluster labels.

        :param X: the input variables as matrix, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: ignored
        :type y: ndarray
        :param sample_weight: Sample weights. If None, then samples are equally weighted. TODO Currently ignored.
        :type sample_weight: array-like of shape (n_samples,), default=None
        :return: the cluster labels (of type int)
        :rtype: ndarray
        """
        self.fit(X)
        return self.predict(X)

    def get_params(self, deep=True):
        """
        Returns the parameters for this cluster, basically classname and options list.

        :param deep: ignored
        :type deep: bool
        :return: the dictionary with options
        :rtype: dict
        """
        result = dict()
        result["classname"] = self._classname
        result["options"] = self._options
        if self._nominal_input_vars is not None:
            result["nominal_input_vars"] = self._nominal_input_vars
        if self._num_nominal_input_labels is not None:
            result["num_nominal_input_labels"] = self._num_nominal_input_labels
        if self._num_nominal_input_labels is not None:
            result["num_nominal_input_labels"] = self._num_nominal_input_labels
        return result

    def set_params(self, **params):
        """
        Sets the options for the cluster, expects 'classname' and 'options'.

        :param params: the parameter dictionary
        :type params: dict
        """
        if len(params) == 0:
            return
        if "classname" not in params:
            raise Exception("Cannot find 'classname' in parameters!")
        if "options" not in params:
            raise Exception("Cannot find 'options' in parameters!")
        self._classname = params["classname"]
        self._options = params["options"]
        self._cluster = Clusterer(classname=self._classname, options=self._options)
        self._nominal_input_vars = None
        if "nominal_input_vars" in params:
            self._nominal_input_vars = params["nominal_input_vars"]
        self._num_nominal_input_labels = None
        if "num_nominal_input_labels" in params:
            self._num_nominal_input_labels = params["num_nominal_input_labels"]

    def __str__(self):
        """
        For printing the model.

        :return: the model representation, if any
        :rtype: str
        """
        if self._cluster is None:
            return self._classname + ": No model built yet"
        else:
            return str(self._cluster)

    def __copy__(self):
        """
        Creates a deep copy of itself.

        :return: the copy
        :rtype: WekaEstimator
        """
        result = WekaCluster(jobject=deepcopy(self.jobject))
        result._classname = self._classname
        result._options = self._options[:]
        return result

    def __repr__(self, N_CHAR_MAX=700):
        """
        Returns a valid Python string using its classname and options.

        :param N_CHAR_MAX: ignored
        :type N_CHAR_MAX: int
        :return: the representation
        :rtype: str
        """
        if isinstance(self._nominal_input_vars, str):
            return "WekaCluster(classname='%s', options=%s, nominal_input_vars='%s')" % (self._cluster.classname, str(self._cluster.options), str(self._nominal_input_vars))
        else:
            return "WekaCluster(classname='%s', options=%s, nominal_input_vars=%s)" % (self._cluster.classname, str(self._cluster.options), str(self._nominal_input_vars))



import numpy as np
from numpy import ndarray
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted, check_array, check_X_y
from weka.filters import Filter
from weka.core.classes import is_instance_of, OptionHandler
from weka.core.dataset import missing_value, Instances
from weka.core.serialization import deepcopy
from sklweka.dataset import to_instances, to_array, to_nominal_attributes, to_nominal_labels


class WekaTransformer(BaseEstimator, OptionHandler, TransformerMixin):
    """
    Wraps a Weka filter within the scikit-learn framework.
    """

    def __init__(self, jobject=None, filter=None, classname=None, options=None,
                 num_nominal_input_labels=None, num_nominal_output_labels=None):
        """
        Initializes the estimator. Can be either instantiated via the following priority of parameters:
        1. JB_Object representing a Java Filter object
        2. Filter pww3 wrapper
        3. classname/options

        :param jobject: the JB_Object representing a Weka filter to use
        :type jobject: JB_Object
        :param filter: the filter wrapper to use
        :type filter: Filter
        :param classname: the classname of the Weka filter to instantiate
        :type classname: str
        :param options: the command-line options of the Weka filter to instantiate
        :type options: list
        :param num_nominal_input_labels: the dictionary with the number of labels for the nominal input variables (key is 0-based attribute index)
        :type num_nominal_input_labels: dict
        :param num_nominal_output_labels: the number of labels for the output variable
        :type num_nominal_output_labels: int
        """
        if jobject is not None:
            _jobject = jobject
        elif filter is not None:
            _jobject = filter.jobject
        elif classname is not None:
            if options is None:
                options = []
            classifier = Filter(classname=classname, options=options)
            _jobject = classifier.jobject
        else:
            raise Exception("At least Java classname must be provided!")

        if not is_instance_of(_jobject, "weka.filters.Filter"):
            raise Exception("Java object does not implement weka.filters.Filter!")

        super(WekaTransformer, self).__init__(_jobject)
        self._filter = Filter(jobject=_jobject)
        self.header_ = None
        # the following references are required for get_params/set_params
        self._classname = classname
        self._options = options
        self._num_nominal_input_labels = num_nominal_input_labels
        self._num_nominal_output_labels = num_nominal_output_labels

    @property
    def filter(self):
        """
        Returns the underlying filter object, if any.

        :return: the classifier object
        :rtype: Classifier
        """
        return self._filter

    @property
    def header(self):
        """
        Returns the underlying dataset header, if any.

        :return: the dataset structure
        :rtype: Instances
        """
        return self.header_

    def fit(self, X, y):
        """
        Trains the estimator.

        :param X: the input variables as matrix, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: the optional class attribute column, array-like of shape (n_samples,)
        :type y: ndarray
        :return: itself
        :rtype: WekaTransformer
        """
        if y is None:
            check_array(X)
        else:
            check_X_y(X, y)
        d = to_instances(X, y=y,
                         num_nominal_labels=self._num_nominal_input_labels,
                         num_class_labels=self._num_nominal_output_labels)
        self.header_ = Instances.template_instances(d)
        self._filter.inputformat(d)
        self._filter.filter(d)
        return self

    def transform(self, X, y=None):
        """
        Filters the data.

        :param X: the data to filter, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: the optional class attribute column, array-like of shape (n_samples,)
        :type y: ndarray
        :return: the filtered data, X if no targets or (X, y) if targets provided
        :rtype: ndarray or tuple
        """
        check_is_fitted(self)
        no_targets = y is None

        # dummy class values necessary?
        if no_targets and self.header_.has_class():
            y = []
            for i in range(X.shape[0]):
                y.append(missing_value())
            y = np.array(y)

        d = to_instances(X, y=y,
                         num_nominal_labels=self._num_nominal_input_labels,
                         num_class_labels=self._num_nominal_output_labels)
        d_new = self._filter.filter(d)
        X, y = to_array(d_new)
        if no_targets:
            return X
        else:
            return X, y

    def get_params(self, deep=True):
        """
        Returns the parameters for this classifier, basically classname and options list.

        :param deep: ignored
        :type deep: bool
        :return: the dictionary with options
        :rtype: dict
        """
        result = dict()
        result["classname"] = self._classname
        result["options"] = self._options
        if self._num_nominal_input_labels is not None:
            result["num_nominal_input_labels"] = self._num_nominal_input_labels
        if self._num_nominal_output_labels is not None:
            result["num_nominal_output_labels"] = self._num_nominal_output_labels
        return result

    def set_params(self, **params):
        """
        Sets the options for the classifier, expects 'classname' and 'options'.

        :param params: the parameter dictionary
        :type params: dict
        """
        if len(params) == 0:
            return
        if "classname" not in params:
            raise Exception("Cannot find 'classname' in parameters!")
        if "options" not in params:
            raise Exception("Cannot find 'options' in parameters!")
        self._classname = params["classname"]
        self._options = params["options"]
        self._filter = Filter(classname=self._classname, options=self._options)
        self._num_nominal_input_labels = None
        if "num_nominal_input_labels" in params:
            self._num_nominal_input_labels = params["num_nominal_input_labels"]
        self._num_nominal_output_labels = None
        if "num_nominal_output_labels" in params:
            self._num_nominal_output_labels = params["num_nominal_output_labels"]

    def __str__(self):
        """
        For printing the model.

        :return: the model representation, if any
        :rtype: str
        """
        if self._filter is None:
            return self._classname + ": No filter instantiated yet"
        else:
            return str(self._filter)

    def __copy__(self):
        """
        Creates a deep copy of itself.

        :return: the copy
        :rtype: WekaTransformer
        """
        result = WekaTransformer(jobject=deepcopy(self.jobject))
        result._classname = self._classname
        result._options = self._options[:]
        return result

    def __repr__(self, N_CHAR_MAX=700):
        """
        Returns a valid Python string using its classname and options.

        :param N_CHAR_MAX: ignored
        :type N_CHAR_MAX: int
        :return: the representation
        :rtype: str
        """
        return "WekaTransformer(classname='%s', options=%s)" % (self._filter.classname, str(self._filter.options))


class MakeNominal(BaseEstimator, TransformerMixin):
    """
    Converts numeric columns to nominal ones (ie string labels).
    """

    def __init__(self, input_vars=None, output_var=False, num_nominal_input_labels=None, num_nominal_output_labels=None):
        """
        Initializes the estimator.

        :param nominal_input_vars: the list of 0-based indices of attributes to convert to nominal or range string with 1-based indices
        :type nominal_input_vars: list or str
        :param output_var: whether to convert the output variable as well
        :type output_var: bool
        :param num_nominal_input_labels: the dictionary with the number of labels for the nominal input variables (key is 0-based attribute index)
        :type num_nominal_input_labels: dict
        :param num_nominal_output_labels: the number of labels for the output variable
        :type num_nominal_output_labels: int
        """
        super(MakeNominal, self).__init__()
        self._input_vars = None if input_vars is None else input_vars[:]
        self._output_var = output_var
        self._num_nominal_input_labels = num_nominal_input_labels
        self._num_nominal_output_labels = num_nominal_output_labels

    @property
    def input_vars(self):
        """
        Returns the 0-based indices or range string with 1-based indices of the input variables to convert.

        :return: the indices or range string, can be None
        :rtype: list or str
        """
        return self._input_vars

    @property
    def output_vars(self):
        """
        Returns whether the output variable gets converted as well.

        :return: True if the output variable gets converted
        :rtype: bool
        """
        return self._input_vars

    def fit(self, X, y):
        """
        Trains the estimator.

        :param X: the input variables as matrix, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: the optional class attribute column, array-like of shape (n_samples,)
        :type y: ndarray
        :return: itself
        :rtype: WekaTransformer
        """
        if y is None:
            check_array(X, dtype=None)
        else:
            check_X_y(X, y, dtype=None)
        self.initialized_ = True
        return self

    def transform(self, X, y=None):
        """
        Filters the data.

        :param X: the data to filter, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: the optional class attribute column, array-like of shape (n_samples,)
        :type y: ndarray
        :return: the filtered data, X if no targets or (X, y) if targets provided
        :rtype: ndarray or tuple
        """
        check_is_fitted(self)
        X_new = to_nominal_attributes(X, self._input_vars)
        y_new = None
        if y is not None:
            y_new = to_nominal_labels(y)
        return X_new, y_new

    def get_params(self, deep=True):
        """
        Returns the parameters for this classifier, basically classname and options list.

        :param deep: ignored
        :type deep: bool
        :return: the dictionary with options
        :rtype: dict
        """
        result = dict()
        result["input_vars"] = self._input_vars
        result["output_var"] = self._output_var
        if self._num_nominal_input_labels is not None:
            result["num_nominal_input_labels"] = self._num_nominal_input_labels
        if self._num_nominal_output_labels is not None:
            result["num_nominal_output_labels"] = self._num_nominal_output_labels
        return result

    def set_params(self, **params):
        """
        Sets the options for the classifier, expects 'classname' and 'options'.

        :param params: the parameter dictionary
        :type params: dict
        """
        if len(params) == 0:
            return
        if "input_vars" not in params:
            raise Exception("Cannot find 'input_vars' in parameters!")
        if "output_var" not in params:
            raise Exception("Cannot find 'output_var' in parameters!")
        self._input_vars = params["input_vars"]
        self._output_var = params["output_var"]
        self._num_nominal_input_labels = None
        if "num_nominal_input_labels" in params:
            self._num_nominal_input_labels = params["num_nominal_input_labels"]
        self._num_nominal_output_labels = None
        if "num_nominal_output_labels" in params:
            self._num_nominal_output_labels = params["num_nominal_output_labels"]

    def __str__(self):
        """
        For printing the model.

        :return: the model representation, if any
        :rtype: str
        """
        return "MakeNominal\n===========\n- Input vars: %s\n- Output var: %s" % (str(self._input_vars), str(self._output_var))

    def __copy__(self):
        """
        Creates a deep copy of itself.

        :return: the copy
        :rtype: WekaTransformer
        """
        return MakeNominal(
            input_vars=(None if self._input_vars is None else self._input_vars[:]),
            output_var=self._output_var)

    def __repr__(self, N_CHAR_MAX=700):
        """
        Returns a valid Python string using its classname and options.

        :param N_CHAR_MAX: ignored
        :type N_CHAR_MAX: int
        :return: the representation
        :rtype: str
        """
        return "MakeNominal(input_vars=%s, output_var=%s)" % (repr(self._input_vars), str(self._output_var))



from setuptools import setup


def _read(f):
    """
    Reads in the content of the file.
    :param f: the file to read
    :type f: str
    :return: the content
    :rtype: str
    """
    return open(f, 'rb').read()


setup(
    name="sklearn-weka-plugin",
    description="Library for making Weka algorithms available within scikit-learn. Relies on the python-weka-wrapper3 library.",
    long_description=(
        _read('DESCRIPTION.rst') + b'\n' +
        _read('CHANGES.rst')).decode('utf-8'),
    url="https://github.com/fracpete/sklearn-weka-plugin",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: Python :: 3',
    ],
    license='GNU General Public License version 3.0 (GPLv3)',
    package_dir={
        '': 'src'
    },
    packages=[
        "sklweka",
    ],
    version="0.0.8",
    author='Peter "fracpete" Reutemann',
    author_email='sklweka@fracpete.org',
    install_requires=[
        "numpy",
        "python-weka-wrapper3>=0.2.5",
        "scikit-learn",
    ],
)



import numpy as np
from numpy import ndarray
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.metrics import accuracy_score, r2_score
from sklweka.preprocessing import to_nominal_attributes, to_nominal_labels
from weka.classifiers import Classifier
from weka.core.classes import is_instance_of, OptionHandler
from weka.core.dataset import missing_value
from weka.core.serialization import deepcopy
from sklweka.dataset import to_instances, to_instance


class WekaEstimator(BaseEstimator, OptionHandler, RegressorMixin, ClassifierMixin):
    """
    Wraps a Weka classifier (classifier/regressor) within the scikit-learn framework.
    """

    def __init__(self, jobject=None, classifier=None, classname=None, options=None,
                 nominal_input_vars=None, nominal_output_var=None,
                 num_nominal_input_labels=None, num_nominal_output_labels=None):
        """
        Initializes the estimator. Can be either instantiated via the following priority of parameters:
        1. JB_Object representing a Java Classifier object
        2. Classifier pww3 wrapper
        3. classname/options

        :param jobject: the JB_Object representing a Weka classifier to use
        :type jobject: JB_Object
        :param classifier: the classifier wrapper to use
        :type classifier: Classifier
        :param classname: the classname of the Weka classifier to instantiate
        :type classname: str
        :param options: the command-line options of the Weka classifier to instantiate
        :type options: list
        :param nominal_input_vars: the list of 0-based indices of attributes to convert to nominal or range string with 1-based indices
        :type nominal_input_vars: list or str
        :param nominal_output_var: whether to convert the output variable to a nominal one
        :type nominal_output_var: bool
        :param num_nominal_input_labels: the dictionary with the number of labels for the nominal input variables (key is 0-based attribute index)
        :type num_nominal_input_labels: dict
        :param num_nominal_output_labels: the number of labels for the output variable
        :type num_nominal_output_labels: int
        """
        if jobject is not None:
            _jobject = jobject
        elif classifier is not None:
            _jobject = classifier.jobject
        elif classname is not None:
            if options is None:
                options = []
            classifier = Classifier(classname=classname, options=options)
            _jobject = classifier.jobject
        else:
            raise Exception("At least Java classname must be provided!")

        if not is_instance_of(_jobject, "weka.classifiers.Classifier"):
            raise Exception("Java object does not implement weka.classifiers.Classifier!")

        super(WekaEstimator, self).__init__(_jobject)
        self._classifier = Classifier(jobject=_jobject)
        self.header_ = None
        self.classes_ = None
        # the following references are required for get_params/set_params
        self._classname = classname if (classname is not None) else self._classifier.classname
        self._options = options if (options is not None) else self._classifier.options
        self._nominal_input_vars = nominal_input_vars
        self._nominal_output_var = nominal_output_var
        self._num_nominal_input_labels = num_nominal_input_labels
        self._num_nominal_output_labels = num_nominal_output_labels

    @property
    def classifier(self):
        """
        Returns the underlying classifier object, if any.

        :return: the classifier object
        :rtype: Classifier
        """
        return self._classifier

    @property
    def header(self):
        """
        Returns the underlying dataset header, if any.

        :return: the dataset structure
        :rtype: Instances
        """
        return self.header_

    def fit(self, X, y, sample_weight=None):
        """
        Trains the estimator.

        :param X: the input variables as matrix, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: the class attribute column, array-like of shape (n_samples,)
        :type y: ndarray
        :param sample_weight: Sample weights. If None, then samples are equally weighted. TODO Currently ignored.
        :type sample_weight: array-like of shape (n_samples,), default=None
        :return: itself
        :rtype: WekaEstimator
        """
        X, y = check_X_y(X, y=y, dtype=None)
        if self._nominal_input_vars is not None:
            X = to_nominal_attributes(X, self._nominal_input_vars)
        if self._nominal_output_var is not None:
            y = to_nominal_labels(y)
        d = to_instances(X, y,
                         num_nominal_labels=self._num_nominal_input_labels,
                         num_class_labels=self._num_nominal_output_labels)
        self._classifier.build_classifier(d)
        self.header_ = d.template_instances(d, 0)
        if d.class_attribute.is_nominal:
            self.classes_ = d.class_attribute.values
        else:
            self.classes_ = None
        return self

    def predict(self, X):
        """
        Performs predictions with the trained classifier.

        :param X: the data matrix to generate predictions for, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :return: the score (or scores)
        :rtype: ndarray
        """
        check_is_fitted(self)
        if self._nominal_input_vars is not None:
            X = to_nominal_attributes(X, self._nominal_input_vars)
        X = check_array(X, dtype=None)
        result = []
        for d in X:
            inst = to_instance(self.header_, d, y=missing_value())
            if self.header_.class_attribute.is_nominal:
                result.append(self.header_.class_attribute.value(int(self._classifier.classify_instance(inst))))
            else:
                result.append(self._classifier.classify_instance(inst))
        return np.array(result)

    def predict_proba(self, X):
        """
        Performs predictions and returns class probabilities.

        :param X: the data matrix to generate predictions for, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :return: the probabilities
        """
        check_is_fitted(self)
        if self._nominal_input_vars is not None:
            X = to_nominal_attributes(X, self._nominal_input_vars)
        X = check_array(X, dtype=None)
        result = []
        for d in X:
            inst = to_instance(self.header_, d, y=missing_value())
            result.append(self._classifier.distribution_for_instance(inst))
        return np.array(result)

    def score(self, X, y, sample_weight=None):
        """
        Classification: return the mean accuracy on the given test data and labels.
        Regression: return the coefficient of determination of the prediction.

        :param X: the input variables as matrix, array-like of shape (n_samples, n_features)
        :type X: ndarray
        :param y: the class attribute column, array-like of shape (n_samples,)
        :type y: ndarray
        :param sample_weight: Sample weights. If None, then samples are equally weighted.
        :type sample_weight: array-like of shape (n_samples,), default=None
        :return: the score
        :rtype: float
        """
        y_pred = self.predict(X)
        if self._nominal_output_var:
            return accuracy_score(y, y_pred, sample_weight=sample_weight)
        else:
            return r2_score(y, y_pred, sample_weight=sample_weight)

    def get_params(self, deep=True):
        """
        Returns the parameters for this classifier, basically classname and options list.

        :param deep: ignored
        :type deep: bool
        :return: the dictionary with options
        :rtype: dict
        """
        result = dict()
        result["classname"] = self._classname
        result["options"] = self._options
        if self._nominal_input_vars is not None:
            result["nominal_input_vars"] = self._nominal_input_vars
        if self._nominal_output_var is not None:
            result["nominal_output_var"] = self._nominal_output_var
        if self._num_nominal_input_labels is not None:
            result["num_nominal_input_labels"] = self._num_nominal_input_labels
        if self._num_nominal_output_labels is not None:
            result["num_nominal_output_labels"] = self._num_nominal_output_labels
        return result

    def set_params(self, **params):
        """
        Sets the options for the classifier, expects 'classname' and 'options'.

        :param params: the parameter dictionary
        :type params: dict
        """
        if len(params) == 0:
            return
        if "classname" not in params:
            raise Exception("Cannot find 'classname' in parameters!")
        if "options" not in params:
            raise Exception("Cannot find 'options' in parameters!")
        self._classname = params["classname"]
        self._options = params["options"]
        self._classifier = Classifier(classname=self._classname, options=self._options)
        self._nominal_input_vars = None
        if "nominal_input_vars" in params:
            self._nominal_input_vars = params["nominal_input_vars"]
        self._nominal_output_var = None
        if "nominal_output_var" in params:
            self._nominal_output_var = params["nominal_output_var"]
        self._num_nominal_input_labels = None
        if "num_nominal_input_labels" in params:
            self._num_nominal_input_labels = params["num_nominal_input_labels"]
        self._num_nominal_output_labels = None
        if "num_nominal_output_labels" in params:
            self._num_nominal_output_labels = params["num_nominal_output_labels"]

    def __str__(self):
        """
        For printing the model.

        :return: the model representation, if any
        :rtype: str
        """
        if self._classifier is None:
            return self._classname + ": No model built yet"
        else:
            return str(self._classifier)

    def __copy__(self):
        """
        Creates a deep copy of itself.

        :return: the copy
        :rtype: WekaEstimator
        """
        result = WekaEstimator(jobject=deepcopy(self.jobject))
        result._classname = self._classname
        result._options = self._options[:]
        result._nominal_input_vars = None if (self._nominal_input_vars is None) else self._nominal_input_vars[:]
        result._nominal_output_var = self._nominal_output_var
        return result

    def __repr__(self, N_CHAR_MAX=700):
        """
        Returns a valid Python string using its classname and options.

        :param N_CHAR_MAX: ignored
        :type N_CHAR_MAX: int
        :return: the representation
        :rtype: str
        """
        if isinstance(self._nominal_input_vars, str):
            return "WekaEstimator(classname='%s', options=%s, nominal_input_vars='%s', nominal_output_var=%s)" % (self._classifier.classname, str(self._classifier.options), str(self._nominal_input_vars), str(self._nominal_output_var))
        else:
            return "WekaEstimator(classname='%s', options=%s, nominal_input_vars=%s, nominal_output_var=%s)" % (self._classifier.classname, str(self._classifier.options), str(self._nominal_input_vars), str(self._nominal_output_var))
